%
% Written by Emanuele Rodola
% TU Munich, 09/2015
%
function [evecs, evals, W, A] = calc_ALB_pcl(points, options)

    function y = psi(x,a)
        %y = a*abs(x);
        y = 1/(1+a*abs(x));
        %y = (1/(1+a*abs(x)))*(x<=0) + (1+a*x)*(x>0);
    end

if ~isfield(options, 'min_angle')
    options.min_angle = pi*15/180;
end
if ~isfield(options, 'radius')
    options.radius = 4*max(range(points))/100;
end
if ~isfield(options, 'warnings')
    options.warnings = false;
end

n = size(points,1);

visualize = false;

if visualize
    figure, plot_mesh(M), hold on; alpha(0.5)
    movegui
end

tic;

% -------------------------------------------------------------------------
% First estimate the normals
% -------------------------------------------------------------------------

normals = zeros(n,3);

for p=1:n
    
    cur_point = points(p,:);
    
    distances = sqrt(sum((repmat(cur_point, n, 1) - points).^2, 2));
    neighs_idx = find(distances>0 & distances<=options.radius);
    neighs = points(neighs_idx,:);
    n_neighs = length(neighs_idx);
    
    barycenter = mean([neighs ; cur_point]);
    
    X = [neighs ; cur_point] - repmat(barycenter,n_neighs+1,1);
    [evecs,evals] = eig(X'*X);
    [~,idx] = sort(diag(evals));
    evecs = evecs(:,idx);
    
    normals(p,:) = evecs(:,1)';
    
end

% figure,plot_mesh(M), shading interp; hold on
% plot_vfield(points,normals,false)
% movegui

% -------------------------------------------------------------------------
% Compute principal curvatures and anisotropic Laplacian
% -------------------------------------------------------------------------

I = [];
J = [];
V = [];
A = zeros(n,1);

curvatures = zeros(n,3);
kmax = zeros(n,1);
kmin = kmax;

ca = cos(-options.angle);
sa = sin(-options.angle);

for p=1:n

    cur_point = points(p,:);
    normal = normals(p,:)';
    
    if visualize
        plot_cloud([], cur_point, 'r.')
        plot_cloud([], cur_point, 'ro')
    end
    
    distances = sqrt(sum((repmat(cur_point, n, 1) - points).^2, 2));
    neighs_idx = find(distances>0 & distances<=options.radius);
    neighs = points(neighs_idx,:);
    n_neighs = length(neighs_idx);
    
    if visualize
        plot_cloud([], neighs, 'b.')
        plot_cloud([], neighs, 'bo')
    end
    
    % Compute principal curvatures
    
    P = eye(3)-normal*normal';
    projected_normals = P*normals(neighs_idx,:)';
    xyz_centroid = mean(projected_normals,2);
    demean_ = projected_normals - repmat(xyz_centroid,1,n_neighs);
    covariance_matrix = demean_*demean_';
    
    [evecs,evals] = eig(covariance_matrix);
    [~,idx] = sort(diag(evals));
    evecs = evecs(:,idx);
    Umax = evecs(:,3)';
    
    % rotate the principal curvature vector by the desired angle
    R = ...
        ca*eye(3) + ...
        sa*[0 -normal(3) normal(2) ; normal(3) 0 -normal(1) ; -normal(2) normal(1) 0] + ...
        (1-ca)*(normal*normal');
    Umax = Umax*R;
    
    curvatures(p,:) = Umax;
    
    kmax(p) = evals(3,3) / n_neighs;
    kmin(p) = evals(2,2) / n_neighs;
    
    % isotropic diffusion
    %D = eye(2);

    % anisotropic diffusion
%     D = [psi(kmin(p),options.alpha) 0 ; 0 psi(kmax(p),options.alpha)];
    D = [1/(1+options.alpha) 0 ; 0 1];
    
    %
    
    barycenter = mean([neighs ; cur_point]);
    
    if visualize
        plot_cloud([], barycenter, 'k*')
    end
    
    X = [neighs ; cur_point] - repmat(barycenter,n_neighs+1,1);
    [evecs,evals] = eig(X'*X);
    [~,idx] = sort(diag(evals));
    evecs = evecs(:,idx);
    
    neighs = neighs - repmat(cur_point,n_neighs,1);
    N_proj_baryc = neighs - (neighs*normal)*normal';
    
    % check projection
    err = norm(N_proj_baryc*normal);
    assert(err < 1e-12)
    
    N_2d = (N_proj_baryc*[evecs(:,2) evecs(:,3)])';
    
    if visualize
        plot_vfield(cur_point,normal',false,'r')
        plot_vfield(cur_point,evecs(:,2)',false)
        plot_vfield(cur_point,evecs(:,3)',false)
    end
    
    % delaunay triangulation
    tris = delaunay([N_2d' ; 0 0]);
    
    % select only the triangle fan around the current point
    [idx,~] = find(tris==size(N_2d,2)+1);
    tri_neigh = sort(tris(idx,:),2);
    
    n_tris = length(idx);
    
    jj = neighs_idx(tri_neigh(:,1:2))';
    
    if n_tris==1
        jj = jj';
        if options.warnings
            warning('Point %d has only one triangle. Consider increasing the neighborhood size.', p)
        end
    end
    
    I = [I ; p*ones(2*n_tris,1)];
    J = [J ; jj(:)];

    jj = [jj ; p*ones(1,n_tris)];
    
    % compute all edges in the triangle fan
    edges = points(jj([2 1 1],:),:) - points(jj([3 3 2],:),:);
    
    % compute triangle normals and area element
    normal_tris = cross(edges(2:3:end,:),edges(1:3:end,:),2); % edge2 x edge1
    norm_tris = sqrt(sum(normal_tris.^2,2));
    normal_tris = normal_tris ./ repmat(norm_tris,1,3);
    A(p) = A(p) + sum(norm_tris)/6;
    
    % normalize the edges and compute all angles in the fan
    edges = edges ./ repmat(sqrt(sum(edges.^2,2)),1,3);
    angles12 = acos(sum(edges(1:3:end,:).*edges(2:3:end,:),2));
    angles23 = acos(sum(edges(2:3:end,:).*edges(3:3:end,:),2));
    angles31 = acos(-sum(edges(1:3:end,:).*edges(3:3:end,:),2));
    
    % compute triangle curvature as the projected avg of its vertices
    Umaxs = reshape(mean(reshape(curvatures(jj,:),3,3*n_tris)),n_tris,3);
    Umaxs = Umaxs - repmat(sum(Umaxs.*normal_tris,2),1,3).*normal_tris;
    Umaxs = Umaxs ./ repmat(sqrt(sum(Umaxs.^2,2)),1,3);
    
    % compute orthogonal curvature direction
    Umins = cross(normal_tris, Umaxs, 2);
    
    if options.warnings
        if any(angles12<options.min_angle | angles23<options.min_angle | angles31<options.min_angle)
            warning('Degenerate triangle(s) detected at point %d.', p)
            %figure, triplot(tris,[N_2d(1,:) 0],[N_2d(2,:) 0]), hold on, plot(0,0,'r*')
        end
    end
    
    % finally compute the stiffness values using FEM
    
    w1s = 0.5 * (...
        D(1,1)*sum(edges(1:3:end,:).*Umins,2).*(-sum(edges(3:3:end,:).*Umins,2)) + ...
        D(2,2)*sum(edges(1:3:end,:).*Umaxs,2).*(-sum(edges(3:3:end,:).*Umaxs,2))) ./ sin(angles31);
    
    w2s = 0.5 * (...
        D(1,1)*sum(edges(2:3:end,:).*Umins,2).*(sum(edges(3:3:end,:).*Umins,2)) + ...
        D(2,2)*sum(edges(2:3:end,:).*Umaxs,2).*(sum(edges(3:3:end,:).*Umaxs,2))) ./ sin(angles23);
    
    %         %w1 = 0.5*cot(angle31);
    %         %w2 = 0.5*cot(angle23);
    %
    %         % p , j1
    %         w1 = 0.5 * (...
    %             D(1,1)*(edge1*Umin')*(-edge3*Umin') + ...
    %             D(2,2)*(edge1*Umax')*(-edge3*Umax')) / sin(angles31(i));
    %
    %         % p , j2
    %         w2 = 0.5 * (...
    %             D(1,1)*(edge2*Umin')*(edge3*Umin') + ...
    %             D(2,2)*(edge2*Umax')*(edge3*Umax')) / sin(angles23(i));
    
    V = [V ; reshape([w1s';w2s'],2*n_tris,1)];
end

% figure, movegui
%
% subplot(1,3,1)
% plot_mesh(M), shading interp; hold on
% plot_vfield(points,curvatures,false)
%
% subplot(1,3,2)
% scatter3(points(:,1),points(:,2),points(:,3),10,kmax,'filled')
% axis equal
% title('kmax')
%
% subplot(1,3,3)
% scatter3(points(:,1),points(:,2),points(:,3),10,kmin,'filled')
% axis equal
% title('kmin')

W = sparse(I, J, V, n, n);

W = 0.5*(W+W');
W = W - sparse(1:n, 1:n, sum(W));

toc

A = sparse(1:n,1:n,A);

[evecs, evals] = eigs(W, A, options.n_eigen, -1e-5, struct('disp', 0));
evals = diag(evals);
evals = abs(evals);
[evals, idx] = sort(evals);
evecs = evecs(:,idx);

end
